# Autogenerated by Astropy-affiliated package astropy_helpers's setup.py on 2017-08-16 14:10:37.888812
from __future__ import unicode_literals
import datetime

version = "2.0.1"
githash = "14ca346b0da3e92e65bdc398cc8f726cbd7be57d"


major = 2
minor = 0
bugfix = 1

release = True
timestamp = datetime.datetime(2017, 8, 16, 14, 10, 37, 888812)
debug = False

try:
    from ._compiler import compiler
except ImportError:
    compiler = "unknown"

try:
    from .cython_version import cython_version
except ImportError:
    cython_version = "unknown"
